package com.marvinelsen.willow

import com.marvinelsen.willow.domain.SearchMode
import com.marvinelsen.willow.domain.SqliteDictionary
import com.marvinelsen.willow.ui.DictionaryEntryFx
import com.marvinelsen.willow.ui.SentenceFx
import com.marvinelsen.willow.ui.toDomain
import com.marvinelsen.willow.ui.toFx
import com.marvinelsen.willow.ui.util.ClipboardHelper
import javafx.beans.property.BooleanProperty
import javafx.beans.property.ObjectProperty
import javafx.beans.property.ReadOnlyBooleanProperty
import javafx.beans.property.ReadOnlyObjectProperty
import javafx.beans.property.SimpleBooleanProperty
import javafx.beans.property.SimpleObjectProperty
import javafx.collections.FXCollections
import javafx.collections.ObservableList
import kotlinx.coroutines.MainScope
import kotlinx.coroutines.launch

class Model(
    private val dictionary: SqliteDictionary,
) {
    private val internalSelectedEntry: ObjectProperty<DictionaryEntryFx> = SimpleObjectProperty()
    private val internalSearchResults: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalWordsBeginning: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalWordsContaining: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalCharacters: ObservableList<DictionaryEntryFx> = FXCollections.observableArrayList()
    private val internalSentences: ObservableList<SentenceFx> = FXCollections.observableArrayList()

    private val internalIsSearching: BooleanProperty = SimpleBooleanProperty(false)
    private val internalIsFindingWordsBeginning: BooleanProperty = SimpleBooleanProperty(false)
    private val internalIsFindingWordsContaining: BooleanProperty = SimpleBooleanProperty(false)
    private val internalIsFindingCharacters: BooleanProperty = SimpleBooleanProperty(false)
    private val internalIsFindingSentences: BooleanProperty = SimpleBooleanProperty(false)

    private val internalFinishedFindingWordsBeginning: BooleanProperty = SimpleBooleanProperty(false)
    private val internalFinishedFindingWordsContaining: BooleanProperty = SimpleBooleanProperty(false)
    private val internalFinishedFindingCharacters: BooleanProperty = SimpleBooleanProperty(false)
    private val internalFinishedFindingSentences: BooleanProperty = SimpleBooleanProperty(false)

    val selectedEntry: ReadOnlyObjectProperty<DictionaryEntryFx> = internalSelectedEntry

    val searchResults: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalSearchResults)
    val wordsBeginning: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalWordsBeginning)
    val wordsContaining: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalWordsContaining)
    val characters: ObservableList<DictionaryEntryFx> =
        FXCollections.unmodifiableObservableList(internalCharacters)
    val sentences: ObservableList<SentenceFx> =
        FXCollections.unmodifiableObservableList(internalSentences)

    val isSearching: ReadOnlyBooleanProperty = internalIsSearching
    val isFindingWordsBeginning: ReadOnlyBooleanProperty = internalIsFindingWordsBeginning
    val isFindingWordsContaining: ReadOnlyBooleanProperty = internalIsFindingWordsContaining
    val isFindingCharacters: ReadOnlyBooleanProperty = internalIsFindingCharacters
    val isFindingSentences: ReadOnlyBooleanProperty = internalIsFindingSentences

    val finishedFindingWordsBeginning: ReadOnlyBooleanProperty = internalFinishedFindingWordsBeginning
    val finishedFindingWordsContaining: ReadOnlyBooleanProperty = internalFinishedFindingWordsContaining
    val finishedFindingCharacters: ReadOnlyBooleanProperty = internalFinishedFindingCharacters
    val finishedFindingSentences: ReadOnlyBooleanProperty = internalFinishedFindingSentences

    private val coroutineScope = MainScope()

    fun search(query: String, searchMode: SearchMode) {
        coroutineScope.launch {
            internalIsSearching.value = true
            internalSearchResults.setAll(dictionary.search(query, searchMode).map { it.toFx() })
            internalIsSearching.value = false
        }
    }

    fun findWordsBeginning() {
        coroutineScope.launch {
            internalIsFindingWordsBeginning.value = true
            internalWordsBeginning.setAll(
                dictionary
                    .findWordsBeginningWith(internalSelectedEntry.value.toDomain())
                    .map { it.toFx() }
            )
            internalIsFindingWordsBeginning.value = false
            internalFinishedFindingWordsBeginning.value = true
        }
    }

    fun findWordsContaining() {
        coroutineScope.launch {
            internalIsFindingWordsContaining.value = true
            internalWordsContaining.setAll(
                dictionary
                    .findWordsContaining(internalSelectedEntry.value.toDomain())
                    .map { it.toFx() }
            )
            internalIsFindingWordsContaining.value = false
            internalFinishedFindingWordsContaining.value = true
        }
    }

    fun findCharacters() {
        coroutineScope.launch {
            internalIsFindingCharacters.value = true
            internalCharacters.setAll(
                dictionary
                    .findCharactersOf(internalSelectedEntry.value.toDomain())
                    .map { it.toFx() }
            )
            internalIsFindingCharacters.value = false
            internalFinishedFindingCharacters.value = true
        }
    }

    fun findSentences() {
        coroutineScope.launch {
            internalIsFindingSentences.value = true
            internalSentences.setAll(
                dictionary
                    .findExampleSentencesFor(internalSelectedEntry.value.toDomain())
                    .map { it.toFx() }
            )
            internalIsFindingSentences.value = false
            internalFinishedFindingSentences.value = true
        }
    }

    fun selectEntry(entry: DictionaryEntryFx) {
        internalWordsBeginning.setAll(emptyList())
        internalWordsContaining.setAll(emptyList())
        internalCharacters.setAll(emptyList())
        internalSentences.setAll(emptyList())

        internalFinishedFindingCharacters.value = false
        internalFinishedFindingWordsBeginning.value = false
        internalFinishedFindingWordsContaining.value = false
        internalFinishedFindingSentences.value = false

        internalSelectedEntry.value = entry
    }

    fun copyHeadwordOfSelectedEntry() {
        ClipboardHelper.copyString(internalSelectedEntry.value.traditionalProperty.value)
    }

    fun copyPronunciationOfSelectedEntry() {
        ClipboardHelper.copyString(internalSelectedEntry.value.pinyinWithToneMarksProperty.value)
    }
}
