package com.marvinelsen.willow.domain

import com.github.houbb.segment.bs.SegmentBs
import com.github.houbb.segment.data.phrase.core.data.SegmentPhraseDatas
import com.github.houbb.segment.data.pos.core.data.SegmentPosDatas
import com.github.houbb.segment.support.format.impl.SegmentFormats
import com.github.houbb.segment.support.segment.impl.Segments
import com.github.houbb.segment.support.segment.mode.impl.SegmentModes
import com.github.houbb.segment.support.segment.result.impl.SegmentResultHandlers
import com.github.houbb.segment.support.tagging.pos.tag.impl.SegmentPosTaggings
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import kotlinx.serialization.json.Json
import java.sql.Connection
import java.sql.ResultSet

class SqliteDictionary(private val connection: Connection) : Dictionary {
    private val whitespaceRegex = """\s+""".toRegex()

    private val dispatcher = Dispatchers.IO
        .limitedParallelism(1)

    private val segmentBs = SegmentBs.newInstance()
        .segment(Segments.defaults())
        .segmentData(SegmentPhraseDatas.define())
        .segmentMode(SegmentModes.dict())
        .segmentFormat(SegmentFormats.chineseSimple())
        .posTagging(SegmentPosTaggings.simple())
        .posData(SegmentPosDatas.define())

    private val searchSimplifiedSql = """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry 
            WHERE simplified GLOB ?
            ORDER BY character_count ASC
    """.trimIndent()

    private val searchTraditionalSql = """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry 
            WHERE traditional GLOB ?
            ORDER BY character_count ASC
    """.trimIndent()

    private val searchPinyinSql = """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry 
            WHERE searchable_pinyin GLOB ?
            OR searchable_pinyin_with_tone_numbers GLOB ?
            ORDER BY character_count ASC
    """.trimIndent()

    private val searchSegmentsSql = """
            WITH cte(id, segment) AS (VALUES ?)
            SELECT entry.traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry INNER JOIN cte
            ON cte.segment = entry.traditional OR cte.segment = entry.simplified
            ORDER BY cte.id
    """.trimIndent()

    private val findWordsBeginningSql = """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry
            WHERE traditional GLOB ?
            ORDER BY character_count ASC
    """.trimIndent()

    private val findWordsContainingSql = """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry
            WHERE traditional LIKE ?
            ORDER BY character_count ASC
    """.trimIndent()

    private val findCharactersSql = """
            WITH cte(id, character, pinyin) AS (VALUES ?)
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, cedict_definitions, cross_straits_definitions, moe_definitions
            FROM entry INNER JOIN cte
            ON cte.character = entry.traditional
            WHERE cte.pinyin = entry.pinyin_with_tone_numbers
            ORDER BY cte.id
    """.trimIndent()

    private val findExampleSentenceSql = """
            SELECT traditional, simplified
            FROM sentence
            WHERE traditional LIKE ?
            ORDER BY character_count ASC
    """.trimIndent()

    override suspend fun search(query: String, searchMode: SearchMode) = withContext(dispatcher) {
        when (searchMode) {
            SearchMode.SIMPLIFIED -> searchSimplified(query)
            SearchMode.TRADITIONAL -> searchTraditional(query)
            SearchMode.PINYIN -> searchPinyin(query)
            SearchMode.SEGMENTS -> searchSegmentsOf(query)
        }
    }

    private fun searchSimplified(simplified: String) =
        connection.prepareStatement(searchSimplifiedSql).use { preparedStatement ->
            preparedStatement.setString(1, "$simplified*")

            preparedStatement.executeQuery().use {
                it.toListOfDictionaryEntries()
            }
        }

    private fun searchTraditional(traditional: String) =
        connection.prepareStatement(searchTraditionalSql).use { preparedStatement ->
            preparedStatement.setString(1, "$traditional*")

            preparedStatement.executeQuery().use {
                it.toListOfDictionaryEntries()
            }
        }

    private fun searchPinyin(pinyin: String): List<DictionaryEntry> {
        val sanitizedPinyin = pinyin.lowercase().replace(whitespaceRegex, "")

        return connection.prepareStatement(searchPinyinSql).use { preparedStatement ->
            preparedStatement.setString(1, "$sanitizedPinyin*")
            preparedStatement.setString(2, "$sanitizedPinyin*")

            preparedStatement.executeQuery().use { resultSet ->
                resultSet.toListOfDictionaryEntries()
            }
        }
    }

    private fun searchSegmentsOf(phrase: String): List<DictionaryEntry> {
        val segments = segmentBs.segment(phrase, SegmentResultHandlers.word())

        val segmentsListString = segments
            .mapIndexed { index, segment -> "($index, '$segment')" }
            .joinToString(",")

        val query = searchSegmentsSql.replace("?", segmentsListString)

        return connection.createStatement().use { statement ->
            statement.executeQuery(query).use { resultSet ->
                resultSet.toListOfDictionaryEntries()
            }
        }
    }

    override suspend fun findWordsContaining(entry: DictionaryEntry) = withContext(dispatcher) {
        connection.prepareStatement(findWordsContainingSql).use { preparedStatement ->
            preparedStatement.setString(1, "_%${entry.traditional}%")

            preparedStatement.executeQuery().use { resultSet ->
                resultSet.toListOfDictionaryEntries()
            }
        }
    }

    override suspend fun findWordsBeginningWith(entry: DictionaryEntry) = withContext(dispatcher) {
        connection.prepareStatement(findWordsBeginningSql).use { preparedStatement ->
            preparedStatement.setString(1, "${entry.traditional}?*")

            preparedStatement.executeQuery().use { resultSet ->
                resultSet.toListOfDictionaryEntries()
            }
        }
    }

    override suspend fun findCharactersOf(entry: DictionaryEntry) = withContext(dispatcher) {
        val pinyinSyllablesWithToneNumbers = entry.pinyinWithToneNumbers
            .lowercase()
            .split(" ")
            .filter { it.isNotBlank() }

        val characters = entry.traditional
            .split("")
            .filter { it.isNotBlank() }
            .filter { it != "，" }
            .filter { it != "·" }

        val charactersWithSyllables = characters.zip(pinyinSyllablesWithToneNumbers)

        val queryInput = charactersWithSyllables
            .mapIndexed { index, s -> "($index, '${s.first}', '${s.second}')" }
            .joinToString(",")

        val query = findCharactersSql.replace("?", queryInput)

        connection.createStatement().use { statement ->
            statement.executeQuery(query).use { resultSet ->
                resultSet.toListOfDictionaryEntries()
            }
        }
    }

    override suspend fun findExampleSentencesFor(entry: DictionaryEntry) = withContext(dispatcher) {
        connection.prepareStatement(findExampleSentenceSql).use { preparedStatement ->
            preparedStatement.setString(1, "_%${entry.traditional}%")

            preparedStatement.executeQuery().use { resultSet ->
                resultSet.toListOfSentences()
            }
        }
    }
}

@Suppress("MagicNumber")
private fun ResultSet.toDictionaryEntry() = DictionaryEntry(
    traditional = this.getString(1),
    simplified = this.getString(2),
    pinyinWithToneMarks = this.getString(3),
    pinyinWithToneNumbers = this.getString(4),
    zhuyin = this.getString(5),
    cedictDefinitions = Json.decodeFromString(this.getString(6)),
    crossStraitsDefinitions = Json.decodeFromString(this.getString(7)),
    moedictDefinitions = Json.decodeFromString(this.getString(8)),
)

private fun ResultSet.toListOfDictionaryEntries() = buildList {
    while (this@toListOfDictionaryEntries.next()) {
        add(this@toListOfDictionaryEntries.toDictionaryEntry())
    }
}

private fun ResultSet.toSentence() = Sentence(
    traditional = this.getString(1),
    simplified = this.getString(2),
)

private fun ResultSet.toListOfSentences() = buildList {
    while (this@toListOfSentences.next()) {
        add(this@toListOfSentences.toSentence())
    }
}
