package com.marvinelsen.willow.ui.controllers

import com.marvinelsen.willow.Model
import com.marvinelsen.willow.config.Config
import com.marvinelsen.willow.config.Pronunciation
import com.marvinelsen.willow.config.Script
import com.marvinelsen.willow.ui.DictionaryEntryFx
import com.marvinelsen.willow.ui.SentenceFx
import com.marvinelsen.willow.ui.cells.DictionaryEntryCellFactory
import com.marvinelsen.willow.ui.cells.SentenceCellFactory
import com.marvinelsen.willow.ui.util.createContextMenuForEntry
import javafx.application.HostServices
import javafx.beans.binding.Bindings
import javafx.fxml.FXML
import javafx.scene.control.Label
import javafx.scene.control.ListView
import javafx.scene.control.ProgressIndicator
import javafx.scene.control.Tab
import javafx.scene.control.TabPane
import javafx.scene.input.ContextMenuEvent
import javafx.scene.layout.FlowPane
import javafx.scene.web.WebView
import kotlinx.html.DIV
import kotlinx.html.body
import kotlinx.html.div
import kotlinx.html.h1
import kotlinx.html.html
import kotlinx.html.li
import kotlinx.html.ol
import kotlinx.html.span
import kotlinx.html.stream.createHTML
import java.util.ResourceBundle

@Suppress("UnusedPrivateMember", "TooManyFunctions")
class DetailsController(
    private val model: Model,
    private val config: Config,
    private val hostServices: HostServices,
) {

    @FXML
    private lateinit var resources: ResourceBundle

    @FXML
    private lateinit var flowPaneHeader: FlowPane

    @FXML
    private lateinit var labelHeadword: Label

    @FXML
    private lateinit var labelPronunciation: Label

    @FXML
    private lateinit var tabPaneDetails: TabPane

    @FXML
    private lateinit var tabCharacters: Tab

    @FXML
    private lateinit var webViewDefinition: WebView

    @FXML
    private lateinit var listViewWordsContaining: ListView<DictionaryEntryFx>

    @FXML
    private lateinit var listViewWordsBeginning: ListView<DictionaryEntryFx>

    @FXML
    private lateinit var listViewCharacters: ListView<DictionaryEntryFx>

    @FXML
    private lateinit var progressIndicatorCharacters: ProgressIndicator

    @FXML
    private lateinit var progressIndicatorWordsContaining: ProgressIndicator

    @FXML
    private lateinit var progressIndicatorWordsBeginning: ProgressIndicator

    @FXML
    private lateinit var labelNoCharactersFound: Label

    @FXML
    private lateinit var labelNoWordsContainingFound: Label

    @FXML
    private lateinit var labelNoWordsBeginningFound: Label

    @FXML
    private lateinit var listViewSentences: ListView<SentenceFx>

    @FXML
    private lateinit var progressIndicatorSentences: ProgressIndicator

    @FXML
    private lateinit var labelNoSentencesFound: Label

    @FXML
    private fun initialize() {
        initializeLabelHeadword()
        initializeLabelPronunciation()
        initializeTabPaneDetails()
        initializeListViewWordsContaining()
        initializeListViewWordsBeginning()
        initializeListViewCharacters()
        initializeListViewSentences()
        initializeWebViewDefinition()
    }

    private fun initializeLabelHeadword() {
        labelHeadword.apply {
            textProperty().bind(
                Bindings.createStringBinding(
                    {
                        val selectedEntry = model.selectedEntry.value
                        when (config.script.value!!) {
                            Script.SIMPLIFIED -> selectedEntry?.simplifiedProperty?.value
                            Script.TRADITIONAL -> selectedEntry?.traditionalProperty?.value
                        }
                    },
                    config.script,
                    model.selectedEntry
                )
            )
            styleProperty().bind(
                Bindings.concat(
                    "-fx-font-size: ",
                    config.details.headwordFontSize.asString(),
                    "px;"
                )
            )
        }
    }

    private fun initializeLabelPronunciation() {
        labelPronunciation.apply {
            textProperty().bind(
                Bindings.createStringBinding(
                    {
                        val selectedEntry = model.selectedEntry.value
                        when (config.details.pronunciation.value!!) {
                            Pronunciation.PINYIN_WITH_TONE_MARKS ->
                                selectedEntry
                                    ?.pinyinWithToneMarksProperty
                                    ?.value

                            Pronunciation.PINYIN_WITH_TONE_NUMBERS ->
                                selectedEntry
                                    ?.pinyinWithToneNumbersProperty
                                    ?.value

                            Pronunciation.ZHUYIN ->
                                selectedEntry
                                    ?.zhuyinProperty
                                    ?.value
                        }
                    },
                    config.details.pronunciation,
                    model.selectedEntry
                )
            )
            styleProperty().bind(
                Bindings.concat(
                    "-fx-font-size: ",
                    config.details.pronunciationFontSize.asString(),
                    "px;"
                )
            )
        }
    }

    private fun initializeTabPaneDetails() {
        tabPaneDetails.apply {
            disableProperty().bind(Bindings.isNull(model.selectedEntry))

            selectionModel.selectedItemProperty().addListener { _, _, selectedTab ->
                if (model.selectedEntry.value == null) return@addListener

                lazyUpdateTabContent(selectedTab.id)
            }
        }

        model.selectedEntry.addListener { _, _, newEntry ->
            if (newEntry == null) return@addListener

            lazyUpdateTabContent(tabPaneDetails.selectionModel.selectedItem.id)
        }

        tabCharacters.disableProperty().bind(
            Bindings.createBooleanBinding(
                {
                    (model.selectedEntry.value?.traditionalProperty?.value?.length ?: 0) < 2
                },
                model.selectedEntry
            )
        )
    }

    private fun initializeListViewSentences() {
        listViewSentences.apply {
            cellFactory = SentenceCellFactory(resources, config, hostServices)
            items = model.sentences

            disableProperty().bind(Bindings.or(model.isFindingSentences, Bindings.isEmpty(model.sentences)))
        }
        progressIndicatorSentences.visibleProperty().bind(model.isFindingSentences)
        labelNoSentencesFound
            .visibleProperty()
            .bind(Bindings.and(Bindings.isEmpty(model.sentences), Bindings.not(model.isFindingSentences)))
    }

    private fun initializeListViewWordsContaining() {
        listViewWordsContaining.apply {
            cellFactory = DictionaryEntryCellFactory(resources, config, hostServices)
            items = model.wordsContaining

            disableProperty().bind(Bindings.or(model.isFindingWordsContaining, Bindings.isEmpty(model.wordsContaining)))
        }
        progressIndicatorWordsContaining.visibleProperty().bind(model.isFindingWordsContaining)
        labelNoWordsContainingFound
            .visibleProperty()
            .bind(Bindings.and(Bindings.isEmpty(model.wordsContaining), Bindings.not(model.isFindingWordsContaining)))
    }

    private fun initializeListViewWordsBeginning() {
        listViewWordsBeginning.apply {
            cellFactory = DictionaryEntryCellFactory(resources, config, hostServices)
            items = model.wordsBeginning

            disableProperty().bind(Bindings.or(model.isFindingWordsBeginning, Bindings.isEmpty(model.wordsBeginning)))
        }
        progressIndicatorWordsBeginning.visibleProperty().bind(model.isFindingWordsBeginning)
        labelNoWordsBeginningFound
            .visibleProperty()
            .bind(Bindings.and(Bindings.isEmpty(model.wordsBeginning), Bindings.not(model.isFindingWordsBeginning)))
    }

    private fun initializeListViewCharacters() {
        listViewCharacters.apply {
            cellFactory = DictionaryEntryCellFactory(resources, config, hostServices)
            items = model.characters

            disableProperty().bind(Bindings.or(model.isFindingCharacters, Bindings.isEmpty(model.characters)))
        }
        progressIndicatorCharacters.visibleProperty().bind(model.isFindingCharacters)
        labelNoCharactersFound
            .visibleProperty()
            .bind(Bindings.and(Bindings.isEmpty(model.characters), Bindings.not(model.isFindingCharacters)))
    }

    private fun initializeWebViewDefinition() {
        webViewDefinition.apply {
            engine.userStyleSheetLocation = this::class.java.getResource("/css/definitions.css")!!.toExternalForm()
        }

        model.selectedEntry.addListener { _, _, newEntry ->
            if (newEntry == null) return@addListener

            webViewDefinition.engine.loadContent(createDefinitionHtml(newEntry))
        }
    }

    @FXML
    private fun headerOnContextMenuRequested(contextMenuEvent: ContextMenuEvent) {
        if (model.selectedEntry.value == null) return

        createContextMenuForEntry(model.selectedEntry.value, resources, hostServices).show(
            flowPaneHeader.scene.window,
            contextMenuEvent.screenX,
            contextMenuEvent.screenY
        )
    }

    @Suppress("ReturnCount")
    private fun lazyUpdateTabContent(selectedTabId: String?) {
        when (selectedTabId) {
            "tabWords" -> {
                if (!model.finishedFindingWordsContaining.value) {
                    model.findWordsContaining()
                }
                if (!model.finishedFindingWordsBeginning.value) {
                    model.findWordsBeginning()
                }
            }

            "tabCharacters" -> {
                if (model.finishedFindingCharacters.value) return

                model.findCharacters()
            }

            "tabSentences" -> {
                if (model.finishedFindingSentences.value) return

                model.findSentences()
            }

            else -> {}
        }
    }

    private fun createDefinitionHtml(entry: DictionaryEntryFx) = createHTML().html {
        body {
            if (entry.cedictDefinitions.isNotEmpty()) {
                div(classes = "cedict-definition") {
                    h1 {
                        +"CC-CEDICT"
                    }
                    cedictDefinition(entry)
                }
            }

            if (entry.crossStraitsDefinitions.isNotEmpty()) {
                div(classes = "cross-straits-definition") {
                    h1 {
                        +"Cross-Straits"
                    }
                    crossStraitsDefinition(entry)
                }
            }

            if (entry.moedictDefinitions.isNotEmpty()) {
                div(classes = "moe-definition") {
                    h1 {
                        +"MOE"
                    }
                    moeDefinition(entry)
                }
            }
        }
    }
}

private fun DIV.cedictDefinition(entry: DictionaryEntryFx) = ol {
    for (definition in entry.cedictDefinitions) {
        li {
            +definition.joinToString(separator = "; ")
        }
    }
}

private fun DIV.crossStraitsDefinition(entry: DictionaryEntryFx) = ol {
    entry.crossStraitsDefinitions.forEach { definition ->
        li {
            span(classes = "definition") {
                +definition.definition
            }
            if (definition.examples.isNotEmpty()) {
                span(classes = "example") {
                    +definition.examples.joinToString(
                        prefix = "如：",
                        separator = "、",
                        postfix = "。"
                    ) { "「$it」" }
                }
            }
        }
    }
}

private fun DIV.moeDefinition(entry: DictionaryEntryFx) =
    entry.moedictDefinitions.groupBy { it.type ?: "" }.entries.forEach { (type, definitions) ->
        if (type != "") {
            span(classes = "type") {
                +type
            }
        }

        ol {
            definitions.forEach { definition ->
                li {
                    span(classes = "definition") {
                        +definition.definition
                    }

                    definition.examples.forEach { example ->
                        span(classes = "example") {
                            +example
                        }
                    }

                    definition.quotes.forEach { quote ->
                        span(classes = "quote") {
                            +quote
                        }
                    }

                    definition.synonyms?.let {
                        span(classes = "synonyms") {
                            +"似：${it.replace(",", "、")}"
                        }
                    }

                    definition.antonyms?.let {
                        span(classes = "antonyms") {
                            +"反：${it.replace(",", "、")}"
                        }
                    }
                }
            }
        }
    }
